"""Test logging library"""
import logging
from cerndb.config import Config
from cerndb.log import setup_root_logger, logger


def test_custom_logger():
    """Test setting a custom logger"""
    # Asser default level
    assert logger.level == logging.INFO  # nosec
    # Change level with Config object
    config = Config()
    config["log"]["level"] = "DEBUG"
    setup_root_logger(config=config, tag="TAG")
    assert logger.level == logging.DEBUG  # nosec
    # Updating config with set_args updates logger
    config.set_args({"log.level": "WARNING"})
    assert logger.level == logging.WARNING  # nosec



"""Module with Basic logging configuration for IT-DB Python modules."""
import logging
from cerndb.config import CFG


class CerndbLoggerException(Exception):
    """Raised when the logger couldn't be initialized."""


# These values are used if the config file is missing them
# or not readable
DEFAULTS = {
    "level": "INFO",
    "file": "/var/log/cerndb.log",
    "fmt": "%(asctime)s [%(module)s:%(lineno)d][%(process)s] " +
           "%(levelname)s {0} %(message)s",
    "datefmt": '%Y-%m-%dT%H:%M:%S%z',
    "console_logging": False,
    "file_logging": True,
}


def setup_root_logger(config=None, tag: str = None):
    """Prepare a root logger.

    You can either use

    CerndbLoggerExceptionRaises:
        CerndbLoggerException:
            - when log_level or log_file are not defined in the config file.
            - when log_level has been incorrectly defined (not one of DEBUG,
                    INFO, WARNING, ERROR, CRITICAL)
    """

    # Auto-reinitialize the logger if --debug provided
    config.register_on_change(setup_root_logger)

    # Fallback to DEFAULTS if the values are not in the config file
    try:
        log_conf = config["log"]
    except (TypeError, KeyError):
        # TypeError: no config defined
        # KeyError: log key not defined in the provided config
        log_conf = DEFAULTS

    log_level = log_conf.get("level", DEFAULTS["level"])
    log_file = log_conf.get("file", DEFAULTS["file"])
    log_to_console = log_conf.get("console_logging",
                                  DEFAULTS['console_logging'])
    log_to_file = log_conf.get("file_logging", DEFAULTS['file_logging'])
    # Prioritize a tag provided via arguments over configuration file and
    # default values
    if tag:
        log_tag = tag
    else:
        log_tag = log_conf.get("tag", "CERNDB")

    if log_level not in ["INFO", "ERROR", "WARNING", "CRITICAL", "DEBUG"]:
        raise CerndbLoggerException(
            f"Incorrect level '{log_level}' in the config "
            f"file {config.used_file}.")

    root_logger = logging.getLogger()
    # To prevent double messages if before importing logging.* was used
    # by default then - basicConfig is run and it adds handlers
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)

    root_logger.setLevel(log_level)

    # Configure formatter
    log_fmt = log_conf.get("fmt", DEFAULTS["fmt"]).format(log_tag)
    log_datefmt = log_conf.get("datefmt", DEFAULTS["datefmt"])
    formatter = logging.Formatter(log_fmt, log_datefmt)

    # Add console logging if enabled. Disabled by default
    if log_to_console or log_level == "DEBUG":
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        root_logger.addHandler(console_handler)

    # Configure file based logger
    if log_to_file:
        try:
            file_handler = logging.FileHandler(log_file)
            file_handler.setFormatter(formatter)
            root_logger.addHandler(file_handler)
        except PermissionError:
            # In nodes where we don't have permissions to write to /var/log
            # we default to a local cerndb.log in the execution folder.
            # This value can then be used or overriden
            # via arguments by the caller
            file_handler = logging.FileHandler('cerndb.log')
            file_handler.setFormatter(formatter)
            root_logger.addHandler(file_handler)


def get_logger(tag: str):
    """Return logger object with default configuration and custom tag"""
    setup_root_logger(config=CFG, tag=tag)
    return logging.getLogger()


setup_root_logger(config=CFG)
logger = logging.getLogger()



"""Test cerndb config."""
from cerndb.config import Config, CFG


def test_default_config():
    """Test default config file is correctly read."""
    config = Config()
    assert config["getpasswd"]["password_directory"] == \
        "/ORA/dbs01/syscontrol/projects/systools/etc"


def test_cfg_contains_default():
    """CFG object should contain default values."""
    assert CFG["log"]["level"] == "INFO"
    assert CFG["getpasswd"]["password_directory"] == \
        "/ORA/dbs01/syscontrol/projects/systools/etc"
    assert isinstance(CFG, Config)


def test_config_initialize_with_dict():
    """Initialize Config object with a dict."""
    con = Config(data={"test": "data"})
    assert con["test"] == "data"



"""get_passwd backends

Raises:
    BackendException: a generic exception
    PasswordDirectoryException: when password_directory doesn't exist
    IOError: when password file exists but cannot be opened from a reason
        different than access denied
    InvalidConfigurationException: missing key in the config file
"""
from abc import ABCMeta, abstractmethod
import os
import errno
import getpass
import argparse
import re
from pathlib import Path
from requests import get
from cerndb.config import CFG, Config
from cerndb.log import logger, setup_root_logger


class InvalidConfigurationException(Exception):
    """Raised when Config file is missing fields."""


class BackendException(Exception):
    """
    Any exception that we expect to happen, and it stops
        further processing. To be caught in main (console runner)
    """


class PasswordDirectoryException(BackendException):
    """
    Raised when specified password_directory doesn't exist
    """


class StatusCodeException(BackendException):
    """
    Raised when GET request returns something else than 200 or 404
    """


class Backend(metaclass=ABCMeta):
    """
    Abstract Backend class, all get_passwd backends should
    inherit from it
    """

    def __init__(self, config=None, settings=None):
        """Read configuration.

        Args:
            config (cerndb.config.Config, optional): Config object to use.
                Defaults to None.
            settings: dict
                Allows to override config file arguments from child class
                    before performing validation
        """
        # User provided or default config
        if config is None:
            self.config = CFG
        else:
            self.config = config
        # Replace config values with what comes from children
        if isinstance(self.config, Config):
            self.config.set_args(settings)
        # validate_config which can be implemented by child classes
        self.validate_config()
        setup_root_logger(self.config,
                          tag="getpasswd.backends.{self.__class__.__name__}")

    def validate_config(self):
        """
        Base implementation of validation. Just check
        if dictionary was given
        """
        if not isinstance(self.config, Config):
            raise InvalidConfigurationException(
                "Incorrect configuration, expected cerndb.config.Config object"
                ", got: {}".format(self.config))

    @classmethod
    def parse_arguments(cls, args):
        """Parse command-line arguments. Child class must implement."""
        if args is not None:
            raise BackendException("Not implemented!")
        return {}

    @abstractmethod
    def get(self, password_tag):
        """Get a password from a password tag."""
        return password_tag


class BackendPasswordFiles(Backend):
    """
    Implementation of the get_passwd backend
    as it was working up until 2019 - reading password files
    located in /ORA/dbs01/syscontrol/projects/systools/etc
    """
    def __init__(self, config=None, password_directory=None):
        """Initialize Backend for files.

        If no arguments are passed, cerndb-base config file is used.
        Otherwise you can either pass a Config object or a password_directory
        param. These 2 are identical:

            config = Config()
            config["getpasswd"]["password_directory"] = "/test"
            BackendPasswordFiles(config=config)
        and
            BackendPasswordFiles(password_directory="/test")



        Args:
            config (cerndb.config.Config, optional): Config object.
                Defaults to None.
            password_directory (str, optional): Directory where to look
                for password files. Defaults to None.

        Raises:
            InvalidConfigurationException: Wrong Config object
            PasswordDirectoryException: Password directory doesn't exist
        """
        settings = {"getpasswd.password_directory": password_directory}
        super().__init__(config=config, settings=settings)

    @classmethod
    def parse_arguments(cls, args):
        """
        Parse command-line arguments specific to this Backend
        """
        parser = argparse.ArgumentParser()
        parser.add_argument("--password_directory",
                            dest="getpasswd.password_directory")
        parser.add_argument(
            "password_tag",
            help="Password tag to retrieve")

        arguments = vars(parser.parse_args(args))

        # If password_directory not provided, try to read env
        if not arguments["getpasswd.password_directory"]:
            arguments["getpasswd.password_directory"] = \
                os.environ.get("GETPASSWD_DIRECTORY", None)
        # Remove not provided arguments, they will come from the config file
        return dict((k, v) for k, v in arguments.items() if v is not None)

    @staticmethod
    def _process_regex(data, tag):
        """
        data - password file data (str)
        tag - password tag to retrieve (str)

        Process the data with a regex, return a list of matching password tags
        """
        regex = r"^TAG {password_tag}=(.*)$"
        regex = regex.format(password_tag=tag)
        pattern = re.compile(regex, re.MULTILINE)

        matches = pattern.findall(data)
        return matches

    def validate_config(self):
        """
        Validations:
            Required keys:
                - password_directory
            - Check password_directory exists
        """
        # call parent implementation
        super().validate_config()
        # Check that we got all the necessary information for this backend
        try:
            password_dir = Path(self.config["getpasswd"]["password_directory"])
        except KeyError as missing_pw_dir:
            raise InvalidConfigurationException(
                "Missing key: password_directory") from missing_pw_dir

        if not password_dir.is_dir():
            raise PasswordDirectoryException(
                f"password_directory doesn't exist: {password_dir}")

    def get(self, password_tag):
        """
        Return password for a given tag

        tag - tag to find in a password file

        returns - password if a tag was found, None if password wasn't found
        """
        # Iterate over file in the passwd directory
        passwd_dir = self.config["getpasswd"]["password_directory"]

        for passwd_file in \
                sorted(os.listdir(passwd_dir), reverse=True):
            # Only filenames starting with passwd
            if not passwd_file.startswith("passwd"):
                continue
            passwd_path = os.path.join(passwd_dir, passwd_file)
            logger.debug("Checking file: %s", passwd_path)
            try:
                with open(passwd_path) as f:
                    # read opened file
                    content = f.read()
            except IOError as e:
                if e.errno == errno.EACCES:
                    # Ignore access denied, just move to the next file
                    logger.debug("Ignoring file %s. Access denied.",
                                 passwd_path)
                    continue
                # Otherwise raise the exception
                raise IOError from e

            matching_passwords = self._process_regex(content, password_tag)
            if matching_passwords:
                user = None
                try:
                    user = getpass.getuser()
                except KeyError:
                    pass
                # Always return the last element, if a password is redefined
                # prefer the latter one
                logger.info("password found in "
                            "PW_FILE=%s "
                            "TAG=%s USER=%s PW_DIR=%s",
                            passwd_file, password_tag,
                            user,
                            passwd_dir)

                return matching_passwords[-1]

        return None


class BackendTeigi(Backend):
    """
    Backend for getting passwords from Teigi
    """
    def validate_config(self):
        """
        Validate if all necessary parameters are in the config file
        or come from command line arguments
        """
        required_keys = [
            "scope",
            "endpoint", "host_private_key_path", "host_certificate_path",
            "cern_ca_bundle_path"]
        if "getpasswd" not in self.config:
            raise InvalidConfigurationException(
                "Missing getpasswd section in the config file")
        # Check getpasswd section of the YAML
        config = self.config["getpasswd"]
        for key in required_keys:
            if key not in config:
                raise InvalidConfigurationException(
                    "Missing key: {}".format(key))
        if 'hostgroup' not in config and config['scope'] == 'hostgroup':
            raise InvalidConfigurationException(
                "Missing key: hostgroup")
        if 'service' not in config and config['scope'] == 'service':
            raise InvalidConfigurationException(
                "Missing key: hostgroup")
        if config['scope'] not in ['service', 'hostgroup', 'hosttree']:
            raise InvalidConfigurationException(
                f"Invalid scope: {config['scope']}")

    @classmethod
    def parse_arguments(cls, args):
        """
        Parse command-line arguments specific to this Backend
        """
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "password_tag",
            help="Password tag to retrieve")
        parser.add_argument(
            "--scope",
            dest="getpasswd.scope",
            choices=["hostgroup", "service", "hosttree"],
            help="Teigi secret scope - hostgroup or service")
        parser.add_argument(
            "--hostgroup",
            dest="getpasswd.hostgroup",
            help="Hostgroup to get the password for")
        parser.add_argument(
            "--service",
            dest="getpasswd.service",
            help="Service to get the password for")
        parser.add_argument(
            "--endpoint",
            dest="getpasswd.endpoint",
            help="Full Teigi URL")
        parser.add_argument(
            "--host_private_key_path",
            dest="getpasswd.host_private_key_path",
            help="Path to host private key")
        parser.add_argument(
            "--host_certificate_path",
            dest="getpasswd.host_certificate_path",
            help="Path to host certificate")
        parser.add_argument(
            "--cern_ca_bundle_path",
            dest="getpasswd.cern_ca_bundle_path",
            help="Path to CERN CA bundle")

        arguments = vars(parser.parse_args(args))

        # Remove not provided arguments, they will come from the config file
        return dict((k, v) for k, v in arguments.items() if v is not None)

    def get(self, password_tag):
        """
        Args:
            password_tag (str): Password tag to retrieve from Teigi

        Get a tag from teigi
        """
        # First, let's format the URL, as it contains a few placeholders
        config = self.config["getpasswd"]
        scope = config["scope"]
        if scope == "hostgroup":
            hostgroup = config["hostgroup"].replace('/', '-')
            endpoint = config["endpoint"].format(
                scope=config["scope"],
                hostgroup=hostgroup,
                password_tag=password_tag)
        elif scope == "service":
            endpoint = config["endpoint"].format(
                scope=config["scope"],
                service=config["service"],
                password_tag=password_tag)
        else:
            endpoint = config["endpoint"].format(
                scope=config["scope"],
                password_tag=password_tag)

        # Call Teigi endpoint with a get request
        logger.debug("Teigi endpoint: %s", endpoint)
        result = get(
            endpoint,
            cert=(
                config['host_certificate_path'],
                config['host_private_key_path']
            ),
            allow_redirects=True,
            verify=config['cern_ca_bundle_path']
        )

        # Check status_code
        #   404 - password not found, return None
        #   200 - password found, return the password
        #   otherwise - raise an exception
        if result.status_code == 200:
            return result.json()["secret"]
        if result.status_code == 404:
            # Password for a given tag not found
            logger.debug("GET status_code: %s", result.status_code)
            return None

        raise(StatusCodeException(f"GET {endpoint},"
                                  f"status code: {result.status_code}"))



# pylint: disable=too-few-public-methods
"""Module for creating SNOW tickets."""
import logging
import json
import requests


class Snow:
    """Class to handle SNOW operations

    This module creates SNOW tickets, the attributes are taken from the
    config object passed
    """
    def __init__(self, config: dict = None):
        """Initialize config object

        Parameters
        ----------
        config: cerndb.config
            A config object for connecting with SNOW
        """
        self.config = config

    def make_incident(self, mail_message: str, mail_title: str):
        """Creates a SNOW Ticket with information about the event

        Parameters
        ----------
        mail_message: str
            The string containing the ticket message
        mail_title: str
            The string containing the title/subject of the ticket
        """
        # Authentication details
        usr = self.config['SNOW_USER']
        psw = self.config['SNOW_PASS']

        # Proxy for https since service-now is hosted outside CERN
        proxies = {
            'https': self.config['SNOW_PROXY']
        }

        # Creating session
        session = requests.Session()
        session.auth = (usr, psw)
        session.proxies.update(proxies)

        # SNOW API Endpoint
        url = f"https://{self.config['SNOW_HOST']}/api/now/v1/table/incident"
        # Timeout (in seconds)
        timeout = int(self.config.get('SNOW_CONNECTION_TIMEOUT_SECONDS', '20'))
        # Standard Headers
        headers = {
            'Accept': 'application/json', 'Content-Type': 'application/json'}

        # Payload for SNOW ticket creation
        data = {
            "short_description": f"{mail_title}",
            "description": f"{mail_message}",
            "comments": f"{mail_message}",
            "u_functional_element": self.config.get(
                'SNOW_DEFAULT_FUNCTIONAL_ELEMENT'),
            "assignment_group": self.config.get(
                'SNOW_DEFAULT_ASSIGNMENT_GROUP'),
            "u_business_service": self.config.get(
                'SNOW_DEFAULT_SERVICE_ELEMENT')
        }

        # Response from SNOW API
        response = session.post(url, headers=headers, data=json.dumps(data),
                                timeout=timeout)

        # Error Handling
        if response.status_code != 201:
            logging.debug("Can't create SNOW ticket : %s", response.json())

        # Get Ticket Number and log it
        incident = response.json()
        ticket_id = incident["result"].get("task_effective_number")
        logging.debug("Ticket ID : %s", ticket_id)

        session.close()



"""
Command line script for get_passwd
"""
# pylint: disable=redefined-builtin
from sys import stdout, exit
import json
import argparse

from cerndb.getpasswd.backends import BackendPasswordFiles, BackendTeigi, \
    InvalidConfigurationException, PasswordDirectoryException
from cerndb.log import logger, setup_root_logger

from cerndb.config import CFG


def parse_arguments():
    """
    Pass command line argument for choosing the backend
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "--backend",
        help="Backend, the default one reads password files",
        dest="getpasswd.backend",
        choices=["files", "teigi"])
    parser.add_argument(
        "--help", "-h",
        help="Display help and exit",
        action="store_true")
    parser.add_argument("--debug", action="store_const",
                        dest="log.level",
                        help="Log level for the Console logger",
                        const="DEBUG")
    parser.add_argument("--verbose", action="store_const",
                        dest="log.level",
                        help="Log level for the Console logger",
                        const="DEBUG")

    # Parse only known, the rest will be parsed in the Backend impl
    known_args, remaining = parser.parse_known_args()
    # Display help from here if no --backend, otherwise, display help from
    # the appropriate backend!
    if known_args.help:
        if getattr(known_args, 'getpasswd.backend'):
            remaining.append("--help")
        else:
            parser.print_help()
            exit(0)
    else:
        known_args.help = None
    # Remove not provided arguments, they will come from the config file
    known_args = dict((k, v) for k, v in vars(known_args).items()
                      if v is not None)
    return (known_args, remaining)


def main():
    """
    This function is invoked by the get_passwd script
    """
    args, remaining = parse_arguments()
    CFG.set_args(args)
    setup_root_logger(CFG, tag="getpasswd.command_line")
    #################################################################
    # Select Backend
    # Read backend from command line, fall back to the config file
    try:
        config = CFG["getpasswd"]
    except KeyError:
        logger.error("Config file %s missing getpasswd section.",
                     CFG.used_file)
        exit(1)

    backend_name = args.get("backend", config.get("backend", None))

    if not backend_name:
        logger.error("Default backend not found in the config file! "
                     "(it can be also specified via --backend)")
        exit(1)

    if backend_name == "files":
        logger.debug("Using PasswordFiles Backend")
        backend_class = BackendPasswordFiles
    elif backend_name == "teigi":
        logger.debug("Using Teigi Backend")
        backend_class = BackendTeigi
    else:
        logger.error("Not supported Backend: %s", backend_name)
        exit(1)
    #################################################################
    # Parse args of a Backend
    arguments = backend_class.parse_arguments(remaining)

    # Create configuration - command line + config file
    CFG.set_args(arguments)

    logger.debug(
        "Used config (config file + command line args): %s",
        json.dumps(config, indent=4, sort_keys=True))
    # Create backend object
    try:
        backend = backend_class(config=CFG)
    except InvalidConfigurationException as exc:
        logger.error("Invalid config file: %s", exc)
        exit(1)
    except PasswordDirectoryException as exc:
        logger.error(exc)
        exit(1)

    password = backend.get(arguments["password_tag"])
    # If password was found - print it
    if password:
        stdout.write(password)
    else:
        # Otherwise exit with 1
        exit(1)


if __name__ == '__main__':
    main()
